/** 
 * \file face_attribute.h
 *  
 * \brief Definition  of the  class FaceAttribute, which  represents a
 * set of face attributes of the underlying triangle surface mesh of a
 * Parametric Pseudo-Surface (PPS) constructed from a Loop subdivision
 * surface.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date January 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef FACE_ATTRIBUTE_H
#define FACE_ATTRIBUTE_H

#include "loop_patch.h"   // LOOPatch


/** 
* \defgroup PPSFROMLOOPNameSpace Namespace ppsfromloop.
* @{
*/

/**
* \namespace ppsfromloop
*
* \brief  The  namespace   ppsfromloop  contains  the  definition  and
* implementation   of   all    classes   representing   a   Parametric
* Pseudo-Surface (PPS) that approximates  the geometry of a given Loop
* subdivision surface.
*/

namespace ppsfromloop {

  /**
   * \class FaceAttribute
   *
   * \brief This class represents a set of attributes of a face of the
   * underlying triangle surface mesh of a PPS constructed from a Loop
   * subdivision surface.
   */
  class FaceAttribute {
  public:
    // ---------------------------------------------------------------
    //
    // Public methods.
    //
    // ---------------------------------------------------------------

    /**
     * \fn FaceAttribute()
     *
     * \brief Creates an instance of this class.
     */
    FaceAttribute() : _patch( 0 )
    {}


    /**
     * \fn FaceAttribute( LOOPatch* patch )
     *
     * \brief Creates an instance of this class.
     *
     * \param patch  A pointer to  the Loop subdivision  surface patch
     * associated with the face that owns this attribute.
     */
    FaceAttribute( LOOPatch* patch ) : _patch( patch )
    {}


    /**
     * \fn FaceAttribute( const FaceAttribute& a )
     *
     * \brief Creates an instance of this class from another instance.
     *
     * \param a A given instance of this class.
     */
    FaceAttribute( const FaceAttribute& a )
    {
      _patch = new LOOPatch( *( a.get_patch() ) ) ;
    }


    /**
     * \fn ~FaceAttribute()
     *
     * \brief Destroys an instance of this class.
     *
     */
    ~FaceAttribute() 
    {
      if ( get_patch() != 0 ) delete get_patch() ;
    }


    /**
     * \fn LOOPatch* get_patch() const
     *
     * \brief  Returns  a  pointer  to the  triangular  surface  patch
     * associated with this face.
     *
     * \return A  pointer to  the triangular surface  patch associated
     * with this face.
     */
    LOOPatch* get_patch() const
    {
      return _patch ; 
    }


    /**
     * \fn void set_patch( LOOPatch* patch )
     *
     * \brief  Assigns an  address to  the pointer  to  the triangular
     * surface patch associated with this face.
     *
     * \param patch The address  of the Loop subdivision surface patch
     * associated with this face.
     */
    void set_patch( LOOPatch* patch )
    {
      _patch = patch ; 
    }

  private:
    // ---------------------------------------------------------------
    //
    // Private data member
    //
    // ---------------------------------------------------------------

    LOOPatch* _patch ;   ///< Pointer to the Loop subdivision surface patch associated with this face. 

  } ;

}

/** @} */ //end of group class.

#endif  // FACE_ATTRIBUTE_H
