/** 
 * \file ppsfrompnt.cpp
 *  
 * \brief  Implementation of  the class  PPSfromPNT that  represents a
 * Parametric Pseudo Surface, which approximates a PN triangle surface
 * defined on a  triangle mesh represented by a  Doubly Connected List
 * (DCEL).
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date January 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include "ppsfrompnt.h"   // PPSfromPNT

#include <cassert>        // assert
#include <cmath>          // fabs, sqrt


/** 
* \defgroup PPSFROMPNTNameSpace Namespace ppsfrompnt.
* @{
*/

/**
* \namespace ppsfrompnt
*
* \brief The namespace ppsfrompnt  contains all classes that implement
* a Parametric  Pseudo-Surface (PPS)  that approximates a  PN triangle
* surface.
*
*/

namespace ppsfrompnt {

  using pps::PPS ;

  /**
   * \fn PPSfromPNT::PPSfromPNT( Surface* mesh )
   *
   * \brief Creates an instance of this PPSfromPNT class. 
   *
   * \param mesh A Pointer to a surface mesh represented by an DCEL.
   */
  PPSfromPNT::PPSfromPNT( Surface* mesh ) : PPS< Surface >( mesh )
  {
    //
    // Sets the owner of each half-edge attribute set.
    //
    for ( EdgeIterator eit = edges_begin() ; !is_done( eit ) ; 
	  move_forward( eit ) ) {
      Edge* edge = get_edge( eit ) ;

      Halfedge* h1 = edge->get_first_halfedge() ;
      Halfedge* h2 = edge->get_second_halfedge() ;

      h1->get_attributes().set_owner( h1 ) ;
      h2->get_attributes().set_owner( h2 ) ; 
    }

    //
    // Create the PN triangle surface.
    //
    build_pnt_surface() ;
  }


  /**
   * \fn void PPSfromPNT::eval_surface( Face* face , double u , double v , double w , double& x , double& y , double& z ) const
   *
   * \brief  Computes a point  on the  PN triangle  patch associated
   * with a given face of the PPS underlying triangle mesh.
   *
   * \param face Pointer to one face of the underlying mesh.
   * \param u First barycentric coordinate of a point on the face.
   * \param v Second barycentric coordinate of a point on the face.
   * \param w Third barycentric coordinate of a point on the face.
   * \param x First Cartesian coordinate of a point on the PPS image.
   * \param y Second Cartesian coordinate of a point on the PPS image.
   * \param z Third Cartesian coordinate of a point on the PPS image.
   */
  void
  PPSfromPNT::eval_surface(
			   Face* face , 
			   double u ,
			   double v ,
			   double w ,
			   double& x ,
			   double& y ,
			   double& z
			  )
    const 
  {
    //
    // Make sure the face has a PN triangle associated with it.
    //
    PNTriangle* patch = face->get_attributes().get_patch() ; 

    assert( patch != 0 ) ;

    //
    // Make sure the coordinates are non-negative and add up to 1.
    //
    assert( 
	   ( u >= 0 ) && ( u <= 1 ) && ( v >= 0 ) && 
	   ( v <= 1 ) && ( w >= 0 ) && ( w <= 1 ) &&
	   ( fabs( 1 - ( u + v + w ) ) <= 1e-15 )
	  ) ;
    //
    // Evaluate the patch.
    //
    patch->point( u , v , x , y , z ) ;
  }


  /**
   * \fn void PPSfromPNT::build_pnt_surface()
   *
   * \brief Computes  a PN triangle surface  patch for each  face of a
   * given triangle mesh, and then assigns the patch to the face.
   *
   */
  void
  PPSfromPNT::build_pnt_surface()
  {
    //
    // For each face of the given triangle mesh, compute a PN triangle. 
    //
    for ( FaceIterator fit = faces_begin() ; !is_done( fit ) ; 
	  move_forward( fit ) ) {
      Face* face = get_face( fit ) ;

      Halfedge* h1 = face->get_halfedge() ;
      Halfedge* h2 = h1->get_next() ;
      Halfedge* h3 = h1->get_prev() ;

      Vertex* v1 = h1->get_origin() ;
      Vertex* v2 = h2->get_origin() ;
      Vertex* v3 = h3->get_origin() ;
      
      double p1[ 3 ] ;
      double p2[ 3 ] ;
      double p3[ 3 ] ;
      
      p1[ 0 ] = v1->x() ;
      p1[ 1 ] = v1->y() ;
      p1[ 2 ] = v1->z() ;
      
      p2[ 0 ] = v2->x() ;
      p2[ 1 ] = v2->y() ;
      p2[ 2 ] = v2->z() ;
      
      p3[ 0 ] = v3->x() ;
      p3[ 1 ] = v3->y() ;
      p3[ 2 ] = v3->z() ;
      
      double n1[ 3 ] ;
      double n2[ 3 ] ;
      double n3[ 3 ] ;
      
      compute_vertex_normal_vector( v1 , n1[ 0 ] , n1[ 1 ] , n1[ 2 ] ) ;
      compute_vertex_normal_vector( v2 , n2[ 0 ] , n2[ 1 ] , n2[ 2 ] ) ;
      compute_vertex_normal_vector( v3 , n3[ 0 ] , n3[ 1 ] , n3[ 2 ] ) ;
      
      //
      // Assign the PN triangle to the face
      //
      face->get_attributes().set_patch( 
	                               new PNTriangle( 
                                                      p1 , 
                                                      p2 , 
                                                      p3 , 
                                                      n1 , 
                                                      n2 , 
                                                      n3 
                                                     ) 
                                      ) ;
    }
    
    return ;
  }


  /**
   * \fn void PPSfromPNT::compute_vertex_normal_vector( Vertex* vertex , double& x , double& y , double& z )
   *
   * \brief Computes a unit vector at a given surface mesh vertex.
   *
   * \param vertex A pointer to a surface mesh vertex.
   * \param x A reference to the first component of the normal vector.
   * \param y A reference to the second component of the normal vector.
   * \param z A reference to the third component of the normal vector.
   */
  void
  PPSfromPNT::compute_vertex_normal_vector(
					   Vertex* vertex ,
					   double& x ,
					   double& y ,
					   double& z
					  )
  {
    unsigned i = 0 ;
    
    x = 0 ;
    y = 0 ;
    z = 0 ;
    
    Halfedge* h1 = vertex->get_halfedge() ;
    Halfedge* h2 = h1 ;
    
    do {
      double xaux ;
      double yaux ;
      double zaux ;
      
      compute_face_normal_vector( h2->get_face() , xaux , yaux , zaux ) ;
      
      x += xaux ;
      y += yaux ;
      z += zaux ;
      
      ++i ;
      
      h2 = h2->get_prev()->get_mate() ;
    }
    while ( h2 != h1 ) ;
    
    x /= double( i ) ;
    y /= double( i ) ;
    z /= double( i ) ;
    
    double length = sqrt( ( x * x ) + ( y * y ) + ( z * z ) ) ;
    
    x /= length ;
    y /= length ;
    z /= length ;  
    
    return ;
  }


  /**
   * \fn void PPSfromPNT::compute_face_normal_vector( Face* face , double& x , double& y , double& z )
   *
   * \brief Computes the unit vector of a given surface mesh face.
   *
   * \param face A pointer to a surface mesh face.
   * \param x A reference to the first component of the normal vector.
   * \param y A reference to the second component of the normal vector.
   * \param z A reference to the third component of the normal vector.
   */
  void
  PPSfromPNT::compute_face_normal_vector( 
					 Face* face , 
					 double& x , 
					 double& y , 
					 double& z 
					)
  {
    Halfedge* h1 = face->get_halfedge() ;
    Halfedge* h2 = h1->get_next() ;
    Halfedge* h3 = h1->get_prev() ;
    
    double x1 = h1->get_origin()->x() ;
    double y1 = h1->get_origin()->y() ;
    double z1 = h1->get_origin()->z() ;

    double x2 = h2->get_origin()->x() ;
    double y2 = h2->get_origin()->y() ;
    double z2 = h2->get_origin()->z() ;
    
    double x3 = h3->get_origin()->x() ;
    double y3 = h3->get_origin()->y() ;
    double z3 = h3->get_origin()->z() ;
    
    x2 -= x1 ;
    y2 -= y1 ;
    z2 -= z1 ;
    
    x3 -= x1 ;
    y3 -= y1 ;
    z3 -= z1 ;
    
    x = y2 * z3 - z2 * y3 ;
    y = z2 * x3 - x2 * z3 ;
    z = x2 * y3 - y2 * x3 ;
    
    double length = sqrt( ( x * x ) + ( y * y ) + ( z * z ) ) ;
    
    x /= length ;
    y /= length ;
    z /= length ;  
    
    return ;
  }

}

/** @} */ //end of group class.
