/** 
 * \file vertex_attribute.h
 *  
 * \brief Definition of the  class VertexAttribute, which represents a
 * set of vertex attributes of the underlying triangle surface mesh of
 * a Parametric  Pseudo-Surface (PPS)  constructed from a  PN triangle
 * surface.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date January 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef VERTEX_ATTRIBUTE_H
#define VERTEX_ATTRIBUTE_H

#include "bezier.h"    // pps::Bezier


/** 
* \defgroup PPSFROMPNTNameSpace Namespace ppsfrompnt.
* @{
*/

/**
* \namespace ppsfrompnt
*
* \brief  The   namespace  ppsfrompnt  contains   the  definition  and
* implementation   of   all    classes   representing   a   Parametric
* Pseudo-Surface (PPS)  that approximates the  geometry of a  given PN
* triangle surface.
*/

namespace ppsfrompnt {

  using pps::Bezier ;


  /**
   * \class VertexAttribute
   *
   * \brief This class  represents a set of attributes  of a vertex of
   * the underlying triangle surface mesh  of a PPS constructed from a
   * PN triangle surface.
   */
  class VertexAttribute {
  public:
    // ---------------------------------------------------------------
    //
    // Public methods.
    //
    // ---------------------------------------------------------------

    /**
     * \fn VertexAttribute()
     *
     * \brief Creates an instance of this class.
     */
    VertexAttribute() : _patch( 0 )
    {}


    /**
     * \fn VertexAttribute( Bezier* patch )
     *
     * \brief Creates an instance of this class.
     *
     * \param  patch  The  address   of  the  PPS  shape  function  (a
     * rectangular  B&eacute;zier patch )  associated with  the vertex
     * that owns this attribute.
     */
    VertexAttribute( Bezier* patch ) : _patch( patch )
    {}


    /**
     * \fn VertexAttribute( const VertexAttribute& a )
     *
     * \brief Creates an instance of this class from another instance.
     *
     * \param a A given instance of this class.
     */
    VertexAttribute( const VertexAttribute& a )
    {
      _patch = new Bezier( *( a.get_patch() ) ) ;
    }


    /**
     * \fn ~VertexAttribute()
     *
     * \brief Destroys an instance of this class.
     *
     */
    ~VertexAttribute() 
    {
      if ( get_patch() != 0 ) delete get_patch() ;
    }


    /**
     * \fn Bezier* get_patch() const
     *
     * \brief  Returns  a  pointer  to  the PPS  shape  function  (  a
     * rectangular  B&eacute;zier patch)  associated  with the  vertex
     * that owns this attribute set.
     *
     * @return  A  pointer  to  the  rectangular  B&eacute;zier  patch
     * associated with the vertex that owns this attribute set.
     */
    Bezier* get_patch() const
    {
      return _patch ; 
    }


    /**
     * \fn void set_patch( Bezier* patch ) 
     *
     * \brief  Assigns  an  address  to  the  patch  pointer  of  this
     * attribute.
     *
     * \param  patch  The  address   of  the  PPS  shape  function  (a
     * rectangular  B&eacute;zier patch )  associated with  the vertex
     * that owns this attribute.
     */
    void set_patch( Bezier* patch )
    {
      _patch = patch ;
    }


  private:
    // ---------------------------------------------------------------
    //
    // Private data member
    //
    // ---------------------------------------------------------------

    Bezier* _patch ;  ///< Pointer to the shape function associated with the vertex that owns this attribute.

  } ;

}

/** @} */ //end of group class.

#endif  // VERTEX_ATTRIBUTE_H
